/*
 * Copyright (c) 2007 Kathryn Huxtable
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * $Id: FilterHandler.java 1155 2007-04-15 01:31:32Z kathryn5 $
 */
package org.kathrynhuxtable.middleware.shibshim.filter;

import java.io.IOException;
import java.io.UnsupportedEncodingException;
import java.util.Date;

import javax.servlet.FilterChain;
import javax.servlet.ServletException;
import javax.servlet.ServletRequest;
import javax.servlet.ServletResponse;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import javax.servlet.http.HttpSession;

import org.apache.log4j.Logger;
import org.kathrynhuxtable.middleware.shibshim.util.AESCrypt;
import org.kathrynhuxtable.middleware.shibshim.util.AESCryptException;
import org.kathrynhuxtable.middleware.shibshim.util.Base64;
import org.kathrynhuxtable.middleware.shibshim.util.RSASignature;
import org.kathrynhuxtable.middleware.shibshim.util.RSASignatureException;

/**
 * Handle an individual Shibboleth Shim filter request. The only reason this
 * class exists is to simplify the calls and to shrink the size of the files.
 * This class is instantiated for each request, so the request and response
 * objects can be instance variables rather than method variables.
 */
public class FilterHandler {
    /**
     * Logger for error/information/debug messages.
     */
    private static Logger       log              = Logger.getLogger(FilterHandler.class.getName());

    /**
     * Default SSL port.
     */
    private static final int    DEFAULT_SSL_PORT = 443;

    /**
     * Saved copy of the request object.
     */
    private HttpServletRequest  request          = null;

    /**
     * Saved copy of the response object.
     */
    private HttpServletResponse response         = null;

    /**
     * Saved copy of the filter chain object.
     */
    private FilterChain         chain            = null;

    /**
     * Saved copy of the filter object.
     */
    private ShibShimFilter      filter           = null;

    /**
     * Saved session object.
     */
    private HttpSession         session          = null;

    /**
     * Saved user attributes object.
     */
    private UserAttributesImpl  attributes       = null;

    /**
     * Instantiate a FilterHandler. Save the request, response, chain, and
     * filter for later use. Parse the application name from the request if not
     * supplied.
     * 
     * @param req
     *            the orginal request object.
     * @param res
     *            the original response object.
     * @param chain
     *            the original filter chain object.
     * @param filter
     *            the original filter object.
     * @throws ImmediateReturnException
     *             if we have already printed a redirect or an error page.
     */
    public FilterHandler(ServletRequest req, ServletResponse res, FilterChain chain, ShibShimFilter filter) throws ImmediateReturnException {
        // The ShibShimFilter is currently defined only for HTTP and HTTPS.
        if (!(req instanceof HttpServletRequest) || !(res instanceof HttpServletResponse)) {
            ErrorPage.displayErrorPage(res, "The Shibboleth Shim filter only supports HTTP and HTTPS");
        }

        // Cast the request and response to the Http forms.
        request = (HttpServletRequest) req;
        response = (HttpServletResponse) res;
        log.debug("ServletPath = \"" + request.getServletPath() + "\"");

        this.chain = chain;
        this.filter = filter;

        // Construct a default application name if none was supplied in the
        // configuration.
        if (filter.getApplication() == null) {
            String application = request.getServerName();
            if (request.getServerPort() != DEFAULT_SSL_PORT) {
                application += ":" + request.getServerPort();
            }
            application += request.getContextPath();
            filter.setApplication(application);
        }

        // Get any cached attributes that may be present.
        session = request.getSession(false);
        if (session != null) {
            attributes = (UserAttributesImpl) session.getAttribute(filter.getSessionAttribute());
        }
    }

    /**
     * <p>
     * Filter requests to make sure user is logged in. If not, redirects to the
     * Shibboleth Shim server, which will log the user in. If logged in, fills
     * in user attributes that can be accessed by the filtered servlet as HTTP
     * headers.
     * </p>
     * 
     * @throws ImmediateReturnException
     *             if we have printed a redirect or an error page.
     */
    public void filterRequest() throws ImmediateReturnException {
        // Check for login/logout requests.
        handleLogin();
        handleLogout();

        // At this point, we know we may require an authenticated session.
        // Create a new session if needed. Get its session ID.
        if (session == null) {
            session = request.getSession(true);
        }

        // Retrieve attributes from session.
        initializeAttributes(false);

        // Check for Assertion Consumer Service redirects.
        // This will retrieve the attributes from the POST redirect from the
        // Shibboleth Shim server.
        handleACS();

        // Make sure that the application and sessionId in
        // the attributes object is current.
        attributes.put("config-application", filter.getApplication());

        // If our user attributes are missing or no longer current, retrieve new
        // values.
        long currentTime = new Date().getTime();
        long verifyTime = getIntConfigHeader("config-verifytime", 0);

        log.debug("Do we need to get user info");
        verifyUser(attributes.getHeader("handle"), currentTime, verifyTime);

        // If we have no session, but one is required, force the user to log in.
        if (!attributes.hasAttributes() && filter.isSessionRequired()) {
            String url = request.getRequestURL().toString();
            String queryString = request.getQueryString();
            if (queryString != null) {
                url += "?" + queryString;
            }

            redirectToLogin(url);
            return;
        }

        // Update time stamps in user attributes.
        if (attributes.hasAttributes()) {
            // Move currentTime into lastTime and set currentTime to the current
            // time.
            long oldCurrentTime = getIntConfigHeader("config-currenttime", currentTime);
            attributes.put("config-lasttime", Long.toString(oldCurrentTime));
            attributes.put("config-currenttime", Long.toString(currentTime));

            // NOPW Production shouldn't need this, especially for apps that get
            // userPassword.
            // log.debug("User info: \"" + attributes + "\"");
        }

        // Redirect if we're handling an ACS, otherwise wrap the request to
        // return the user attributes and chain down the filters.
        if (checkForACSPath()) {
            redirectToTarget();
        } else {
            wrapRequest();
        }
    }

    /**
     * We've just logged in and set the attributes. Redirect to the original URL
     * that the user wanted.
     * 
     * @throws ImmediateReturnException
     *             because we have already printed a redirect.
     */
    private void redirectToTarget() throws ImmediateReturnException {
        String url = request.getParameter("target");
        if (url == null) {
            url = "";
        } else if (url.charAt(0) != '/') {
            url = request.getContextPath() + "/" + url;
        } else {
            url = request.getContextPath() + url;
        }
        generalShibShimRedirect(url);
    }

    /**
     * Wrap the request to return principal attributes and pass the request and
     * response along the filter chain.
     * 
     * @throws ImmediateReturnException
     *             if we have printed an error page.
     */
    private void wrapRequest() throws ImmediateReturnException {
        RequestWrapper requestWrapper = new RequestWrapper(request, attributes, filter.getRemoteUserAttribute());

        try {
            chain.doFilter(requestWrapper, response);
        } catch (ServletException se) {
            ErrorPage.displayErrorPage(response, "Error protecting application: " + se);
        } catch (IOException ie) {
            ErrorPage.displayErrorPage(response, "Error protecting application: " + ie);
        }
    }

    /**
     * Is the request for the local Assertion Consumer Service?
     * 
     * @return true if the URL is for the local Assertion Consumer Service.
     */
    private boolean checkForACSPath() {
        return request.getServletPath().equals(filter.getLocalAcsPath());
    }

    /**
     * Parse the attributes from the XML document and verify the signature.
     * 
     * @throws ImmediateReturnException
     *             if we print an error page.
     */
    private void handleACS() throws ImmediateReturnException {
        if (!checkForACSPath()) {
            return;
        }

        attributes.clearUserInfo();
        log.debug("Trying to get user info");
        try {
            String userInfo = extractAndVerifyUserInfo();
            attributes.parseUserInfo(userInfo);
            attributes.put("config-attributes", userInfo);
        } catch (UserInfoException e) {
            ErrorPage.displayErrorPage(response, e.getMessage());
        }
        log.debug("Back from getting user info");
    }

    /**
     * Is the request for the login path, requiring the user to log in?
     * 
     * @return true if the request is for the login path.
     */
    private boolean checkForLoginPath() {
        return request.getServletPath().equals(filter.getLoginRedirectPath());
    }

    /**
     * Process forced login. Make sure there is a context session and session
     * attributes, and clear the session attribute data, if any before
     * redirecting to the Shibboleth Shim server.
     * 
     * @throws ImmediateReturnException
     *             if we print a redirect or an error page.
     */
    private void handleLogin() throws ImmediateReturnException {
        if (!checkForLoginPath()) {
            return;
        }

        // Make sure we have a session established and clear the attributes.
        if (session == null) {
            session = request.getSession(true);
        }
        initializeAttributes(true);

        // Get the desired target URL.
        String target = request.getParameter("target");
        if (target == null) {
            target = filter.getApplication();
        } else {
            try {
                target = java.net.URLDecoder.decode(target, "US-ASCII");
            } catch (IOException ie) {
                log.error("IOException encountered decoding target: " + ie);
                target = filter.getApplication();
            }
        }

        // Redirect to the Shibboleth Shim filter to log the user in if
        // necessary.
        String force = request.getParameter("force");
        if (force != null && force.equalsIgnoreCase("true")) {
            redirectToForceLogin(session.getId(), target);
        } else {
            redirectToLogin(target);
        }
    }

    /**
     * Is the request for the logout path, forcing a logout?
     * 
     * @return true if the request is for the logout path.
     */
    private boolean checkForLogoutPath() {
        return request.getServletPath().equals(filter.getLogoutRedirectPath());
    }

    /**
     * Process logout. Invalidate the servlet session, if any, and redirect to
     * ShibShim logout, which will destroy the main session data. Note that
     * invalidating the context session will also destroy the ShibShim session.
     * 
     * @throws ImmediateReturnException
     *             if we print a redirect.
     */
    private void handleLogout() throws ImmediateReturnException {
        if (!checkForLogoutPath()) {
            return;
        }

        // Invalidate the server session.
        if (session != null) {
            session.invalidate();
        }

        // Redirect to the Shibboleth Shim server, specifying a target URL.
        String target = request.getParameter("target");
        if (target != null) {
            try {
                target = java.net.URLDecoder.decode(target, "US-ASCII");
            } catch (IOException ie) {
                log.error("IOException encountered decoding target: " + ie);
                target = "";
            }
        }
        redirectToLogout(target);
    }

    /**
     * Is it time to verify with the Shibboleth Shim server that the user's
     * session is still logged in?
     * 
     * @param currentTime
     *            the current time.
     * @param verifyTime
     *            the last time the session was verified.
     * @return true if it is time to verify the session.
     */
    private boolean isTimeToVerifyAuthentication(long currentTime, long verifyTime) {
        // TODO need to verify session
        // The check is
        // (!attributes.hasAttributes() || verifyTime == 0 || currentTime - verifyTime >= filter.getTimeout())
        return false;
    }

    /**
     * Verify the session. At the moment this does nothing.
     * 
     * @param handle
     *            the session handle for the Shibboleth Shim server.
     * @param currentTime
     *            the current time.
     * @param verifyTime
     *            the last time the session was verified.
     * @throws ImmediateReturnException
     *             if we print a redirect.
     */
    private void verifyUser(String handle, long currentTime, long verifyTime) throws ImmediateReturnException {
        if (!isTimeToVerifyAuthentication(currentTime, verifyTime)) {
            return;
        }
    }

    /**
     * Redirects browser to the Shibboleth Shim server to perform a login.
     * 
     * @param url
     *            the URL to redirect to after login/out.
     * 
     * @throws ImmediateReturnException
     *             if we print a redirect or an error page.
     */
    protected void redirectToLogin(String url) throws ImmediateReturnException {
        String acs = filter.getAcsUrl();
        int start = 0;
        if (url.startsWith("https://")) {
            acs += "/https/";
            start = "https://".length();
        } else if (url.startsWith("http://")) {
            acs += "/http/";
            start = "http://".length();
        }

        int slash = url.indexOf('/', start);
        if (slash < 0) {
            slash = url.length();
        }
        acs += url.substring(start, slash) + request.getContextPath() + filter.getLocalAcsPath();

        String target = url.replaceFirst("^https?://[^/]+" + request.getContextPath(), "");

        // The url https://myhost.example.edu/foo/protected/index.jsp yields:
        //
        // acs = /https/myhost.example.edu/foo/localACS
        //
        // target = /protected/index.jsp

        log.debug("User info is still empty");
        try {
            String redirectUrl = acs + "?target=" + java.net.URLEncoder.encode(target, "US-ASCII");
            generalShibShimRedirect(redirectUrl);
        } catch (UnsupportedEncodingException e) {
            ErrorPage.displayErrorPage(response, "Got error attempting to log in");
        }
    }

    /**
     * Redirects browser to the Shibboleth Shim server and force login.
     * @param sessionId
     *            the session ID to invalidate on logout.
     * @param returnUrl
     *            the URL to redirect to after login/out.
     * 
     * @throws ImmediateReturnException
     *             if we print a redirect or an error page.
     */
    protected void redirectToForceLogin(String sessionId, String returnUrl) throws ImmediateReturnException {
        try {
            String url = filter.getAcsUrl() + "?" + "return=" + java.net.URLEncoder.encode(returnUrl, "US-ASCII");
            generalShibShimRedirect(url);
        } catch (UnsupportedEncodingException e) {
            ErrorPage.displayErrorPage(response, "Got error attempting to log in");
        }
    }

    /**
     * Redirects browser to Shibboleth Shim server to force a logout.
     * @param returnUrl
     *            the URL to redirect to after login/out.
     * 
     * @throws ImmediateReturnException
     *             if we print a redirect or an error page.
     */
    protected void redirectToLogout(String returnUrl) throws ImmediateReturnException {
        try {
            String url = filter.getAcsUrl();
            if (returnUrl != null) {
                url += "?return=" + java.net.URLEncoder.encode(returnUrl, "US-ASCII");
            }
            generalShibShimRedirect(url);
        } catch (UnsupportedEncodingException e) {
            ErrorPage.displayErrorPage(response, "Got error attempting to log out");
        }
    }

    /**
     * Redirects browser. Used by the other redirect methods.
     * @param url
     *            the URL to redirect to after login/out.
     * 
     * @throws ImmediateReturnException
     *             because we printed a redirect.
     */
    protected void generalShibShimRedirect(String url) throws ImmediateReturnException {
        try {
            log.debug("Redirecting to \"" + url + "\"");
            response.sendRedirect(url);
        } catch (IOException e) {
            // Do nothing.
        }
        throw new ImmediateReturnException();
    }

    /**
     * Extract the XML document from the Base64 encoded assertion and verify the
     * signature in the Base64 encoded signature.
     * 
     * @return the XML attribute assertion.
     * @throws ImmediateReturnException
     *             if we display an error page.
     */
    protected String extractAndVerifyUserInfo() throws ImmediateReturnException {
        // Retrieve the assertion and decrypt it.
        byte[] clearText = null;
        try {
            String base64CipherText = request.getParameter("assertion");
            byte[] cipherText = Base64.decode(base64CipherText);
            clearText = AESCrypt.decrypt(cipherText, filter.getShibShimServerCryptKey());
        } catch (AESCryptException e) {
            ErrorPage.displayErrorPage(response, "AES decrypt exception: " + e);
        }

        // Get the signature and verify the assertion against it.
        boolean status = false;
        try {
            String signature = request.getParameter("signature");
            byte[] sigBlock = Base64.decode(signature);
            status = RSASignature.verify(clearText, sigBlock, filter.getShibShimServerCert());
        } catch (RSASignatureException e) {
            ErrorPage.displayErrorPage(response, "RSA signature exception: " + e);
        }

        if (!status) {
            ErrorPage.displayErrorPage(response, "The Shibboleth Shim Server signature failed to verify");
        }

        String assertion = new String(clearText);
        log.debug("assertion = \"" + assertion + "\"");
        return assertion;
    }

    /**
     * Get the value for an integer header in the configuration attributes.
     * 
     * @param attr
     *            the attribute name.
     * @param defaultValue
     *            the value to return if the attribute is not present.
     * @return the value of the attribute, or the defaultValue if not present.
     * @throws ImmediateReturnException
     *             if we print an error page because the value did not parse as
     *             an integer.
     */
    private long getIntConfigHeader(String attr, long defaultValue) throws ImmediateReturnException {
        long intValue = defaultValue;
        String tmp = (String) attributes.get(attr);
        if (tmp != null) {
            try {
                intValue = Long.parseLong(tmp);
            } catch (NumberFormatException e) {
                ErrorPage.displayErrorPage(response, e.getMessage());
            }
        }
        return intValue;
    }

    /**
     * If necessary, create the attributes object and initialize it.
     * 
     * @param clearAttributes
     *            set to <tt>true</tt> if the attributes from the session
     *            object should be cleared.
     * @throws ImmediateReturnException
     *             if we print an error page.
     */
    private void initializeAttributes(boolean clearAttributes) throws ImmediateReturnException {
        if (attributes == null) {
            attributes = new UserAttributesImpl(filter.getAttributeMap());
            if (attributes == null) {
                log.error("Cannot create " + filter.getSessionAttribute() + " object");
                ErrorPage.displayErrorPage(response, "Unable to log in due to a system problem on the application's server");
            }
            session.setAttribute(filter.getSessionAttribute(), attributes);
        }

        if (clearAttributes) {
            attributes.clearUserInfo();
        }
    }
}
