/*
 * Copyright (c) 2007 Kathryn Huxtable
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * $Id: ShibShimFilter.java 1146 2007-04-14 10:28:25Z kathryn5 $
 */
package org.kathrynhuxtable.middleware.shibshim.filter;

import java.io.BufferedReader;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStreamReader;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.HashMap;
import java.util.Map;

import javax.servlet.Filter;
import javax.servlet.FilterChain;
import javax.servlet.FilterConfig;
import javax.servlet.ServletException;
import javax.servlet.ServletRequest;
import javax.servlet.ServletResponse;

import org.apache.log4j.Logger;
import org.kathrynhuxtable.middleware.shibshim.util.properties.PropertiesBundle;
import org.kathrynhuxtable.middleware.shibshim.util.properties.PropertiesBundleException;

/**
 * The servlet filter for the Shibboleth Shim.
 */
public final class ShibShimFilter implements Filter {
    /**
     * Logger for error/information/debug messages.
     */
    private static Logger    log             = Logger.getLogger(ShibShimFilter.class.getName());

    /**
     * Default timeout value.
     */
    private static final int DEFAULT_TIMEOUT = 60;

    /**
     * Number of milliseconds per second.
     */
    private static final int MILLI_PER_SEC   = 1000;

    /**
     * The URL for the Shibboleth Shim server Assertion Consumer Service.
     */
    private String           acsUrl;

    /**
     * The application name.
     */
    private String           application;

    /**
     * The Map mapping incoming attribute names to the attributes stored in the
     * UserAttribute object.
     */
    private Map              attributeMap    = new HashMap();

    /**
     * The context-relative path for the local Shibboleth Shim filter Assertion
     * Consumer Service.
     */
    private String           localAcsPath;

    /**
     * The context-relative path for the login redirect URL.
     */
    private String           loginRedirectPath;

    /**
     * The context-relative path for the logout redirect URL.
     */
    private String           logoutRedirectPath;

    /**
     * The name of the Shibboleth attribute that will be mapped to the remote
     * user such that request.getRemoteUser() will return it.
     */
    private String           remoteUserAttribute;

    /**
     * The name to use to store the session attribute containing the
     * UserAttributes object.
     */
    private String           sessionAttribute;

    /**
     * Do we require a session, or can the user access the path without logging
     * in? Set to <tt>false</tt> if login is not required. The default is
     * <tt>true</tt>.
     */
    private boolean          sessionRequired;

    /**
     * A URL designating the file holding the Shibboleth Shim server's signing
     * certificate.
     */
    private String           shibShimServerCert;

    /**
     * A URL designating the file holding the Shibboleth Shim server's encryption
     * key.
     */
    private String           shibShimServerCryptKey;

    /**
     * The number of seconds permitted before the session must be verified with
     * the Shibboleth Shim server. Not currently used.
     */
    private long             timeout;

    /**
     * @return the acsUrl
     */
    public String getAcsUrl() {
        return acsUrl;
    }

    /**
     * @return the application
     */
    public String getApplication() {
        return application;
    }

    /**
     * @param application
     *            the application to set
     */
    public void setApplication(String application) {
        this.application = application;
    }

    /**
     * @return the attributeMap
     */
    public Map getAttributeMap() {
        return attributeMap;
    }

    /**
     * @return the localAcsPath
     */
    public String getLocalAcsPath() {
        return localAcsPath;
    }

    /**
     * @return the loginRedirectPath
     */
    public String getLoginRedirectPath() {
        return loginRedirectPath;
    }

    /**
     * @return the logoutRedirectPath
     */
    public String getLogoutRedirectPath() {
        return logoutRedirectPath;
    }

    /**
     * @return the remoteUserAttribute
     */
    public String getRemoteUserAttribute() {
        return remoteUserAttribute;
    }

    /**
     * @return the sessionAttribute
     */
    public String getSessionAttribute() {
        return sessionAttribute;
    }

    /**
     * @return the sessionRequired
     */
    public boolean isSessionRequired() {
        return sessionRequired;
    }

    /**
     * @return the shibShimServerCert
     */
    public String getShibShimServerCert() {
        return shibShimServerCert;
    }

    /**
     * @return the shibShimServerCryptKey
     */
    public String getShibShimServerCryptKey() {
        return shibShimServerCryptKey;
    }

    /**
     * @return the timeout
     */
    public long getTimeout() {
        return timeout;
    }

    /**
     * Saves the config object. Initializes any private attributes from
     * init-param values.
     * 
     * @param config
     *            the configuration object for this filter.
     * 
     * @throws ServletException
     *             if the properties aren't present or there is no acsURL
     *             property.
     */
    public void init(FilterConfig config) throws ServletException {
        String propertiesFile = config.getInitParameter("ShibShimFilterPropertiesFile");
        if (propertiesFile == null || propertiesFile.length() <= 0) {
            log.fatal("No Shibboleth Shim filter properties file specified in filter parameters.");
            log.fatal("Use filter param ShibShimFilterPropertiesFile.");
            throw new ServletException("Incorrect configuration data supplied.");
        }

        // Load the properties bundle from the properties file.
        // Get the values.
        try {
            PropertiesBundle.init(propertiesFile);

            // Get the required attributes
            acsUrl = PropertiesBundle.getString("Assertion Consumer Service URL", "acsUrl");
            attributeMap = PropertiesBundle.getMap("attribute map", "attributeMap.");
            shibShimServerCert = readFile(PropertiesBundle.getString("server cert file", "certFile"));
            shibShimServerCryptKey = readFile(PropertiesBundle.getString("server encryption key file", "cryptKeyFile"));
            localAcsPath = PropertiesBundle.getString("local Assertion Consumer Service path", "localAcsPath");
            sessionAttribute = PropertiesBundle.getString("name of session attribute", "sessionAttribute");

            // Get the optional attributes.
            application = PropertiesBundle.getString("application", "application", null);
            loginRedirectPath = PropertiesBundle.getString("login redirect path", "loginRedirectPath", null);
            logoutRedirectPath = PropertiesBundle.getString("logout redirect path", "logoutRedirectPath", null);
            remoteUserAttribute = PropertiesBundle.getString("remote user attribute", "remoteUserAttribute", null);
            sessionRequired = "true".equals(PropertiesBundle.getString("session required", "sessionRequired", "true"));
            timeout = PropertiesBundle.getInteger("session timeout", "timeout", DEFAULT_TIMEOUT) * MILLI_PER_SEC;
        } catch (PropertiesBundleException e) {
            throw new ServletException(e);
        }

        log.debug("ShibShimFilter initialized");
    }

    /**
     * Overrides the destroy method from filter.
     */
    public void destroy() {
        // Do nothing.
    }

    /**
     * <p>
     * Filters requests to make sure user is logged in. If not, redirects to
     * standard login page. If logged in, fills in some principal attributes
     * that can be used by the filtered servlet.
     * </p>
     * 
     * <p>
     * Note: The Filter will throw a Servlet Exception only in one catastrophic
     * case, where it is not running in a Web Server?!? It doesn't generate
     * other exceptions, but will pass on exceptions generated by other Filters
     * or the Servlet farther down the chain.
     * </p>
     * 
     * @param req
     *            the servlet request object.
     * @param res
     *            the servlet response object.
     * 
     * @param chain
     *            the chain of servlets to run if we are logged in.
     * 
     * @throws IOException
     * @throws ServletException
     */
    public void doFilter(ServletRequest req, ServletResponse res, FilterChain chain) {
        try {
            FilterHandler handler = new FilterHandler(req, res, chain, this);
            handler.filterRequest();
        } catch (ImmediateReturnException e) {
            // Do nothing. We have already redirected to another page or printed
            // an error page.
        }
    }

    /**
     * Read a file and return its text as a string.
     * 
     * @param filename
     *            the name of the file.
     * @return the text of the file.
     * @throws ServletException
     *             if the file cannot be read.
     */
    private String readFile(String filename) throws ServletException {
        String text = null;
        try {
            URL resource = null;
            try {
                resource = new URL(filename);
            } catch (MalformedURLException e) {
                resource = PropertiesBundle.class.getResource(filename);
            }
            if (resource == null) {
                log.fatal("File could not be found at the specified location: " + filename);
                throw new ServletException("File could not be found at the specified location: " + filename);
            }
            BufferedReader reader = new BufferedReader(new InputStreamReader(resource.openStream()));
            StringBuffer textBuffer = new StringBuffer();
            String buffer = null;
            while ((buffer = reader.readLine()) != null) {
                textBuffer.append(buffer).append('\n');
            }
            reader.close();
            text = new String(textBuffer);
        } catch (FileNotFoundException e) {
            throw new ServletException("File not found: " + filename, e);
        } catch (IOException e) {
            throw new ServletException("IOException reading file: " + filename, e);
        }
        return text;
    }
}
